#include "mmc_lib.h"
#include "inttypes.h"
#include "vfat.h"

static unsigned char fat_buf[512];
static uint32_t currentfatsector;

char str[20];

volatile uint16_t  RootDirectoryRegionStart;
volatile uint32_t  DataRegionStart;
volatile uint16_t  FATRegionSize;
volatile uint16_t  RootDirectoryRegionSize;
volatile uint32_t  DataRegionSize;
volatile uint8_t   SectorsPerCluster;
volatile uint16_t  BytePerSector;
volatile uint16_t  FATRegionStart;
volatile uint16_t  NumberOfFATs;
volatile uint32_t  SectorsPerFAT;
volatile uint32_t  VolumeStart;
volatile uint16_t  ReservedSectors;

uint8_t partitiontype;

unsigned char vfat_init(void)
{
	mbr_t *mbr = (mbr_t*) fat_buf;
	vbr_t *vbr = (vbr_t*) fat_buf;
	unsigned char i;
	
	if (mmc_init() != MMC_OK)
		return 0;
	
	//Load MBR
	mmc_set_blocklen(512);
	mmc_start_read_block(0, 9);
	mmc_read_buffer((unsigned char *) mbr, 512);
	mmc_stop_read_block();
	
	for (i=0; i<4; i++)
	{
		if ((mbr->sector.partition[i].typeId == PARTTYPE_FAT12) ||
			 (mbr->sector.partition[i].typeId == PARTTYPE_FAT16L) ||
			 (mbr->sector.partition[i].typeId == PARTTYPE_FAT32))
			break;
	}	
	
	if (i > 3)
		return 0;
	
	partitiontype = mbr->sector.partition[i].typeId;
		
	VolumeStart = mbr->sector.partition[i].sectorOffset;
	
	//Load VBR
	mmc_start_read_block(mbr->sector.partition[i].sectorOffset, 9);
	mmc_read_buffer((unsigned char *) vbr, 512);
	mmc_stop_read_block();
	
	ReservedSectors 		    = vbr->bsRsvdSecCnt; 		// 6
	BytePerSector 				= vbr->bsBytesPerSec;		// 512
	SectorsPerCluster  			= vbr->bsSecPerClus;		// 4
	NumberOfFATs 		 		= vbr->bsNumFATs;			// 2
	
	if (partitiontype == PARTTYPE_FAT32)
		SectorsPerFAT = vbr->SecPerFAT32;
	else
		SectorsPerFAT 		 		= vbr->bsNrSeProFAT16;		// 245
	
	// Calculation Algorithms
	FATRegionStart				= VolumeStart + ReservedSectors;						// 6
	RootDirectoryRegionStart 	= FATRegionStart + (NumberOfFATs * SectorsPerFAT);		// 496	
	RootDirectoryRegionSize 	= (vbr->bsRootEntCnt * 32)/ BytePerSector;					// 32
	DataRegionStart 			= RootDirectoryRegionStart + RootDirectoryRegionSize;	// 528
	FATRegionSize 				= NumberOfFATs * SectorsPerFAT;							// 490
	DataRegionSize				= vbr->bsTotSec32 - (ReservedSectors + FATRegionSize + RootDirectoryRegionSize);	
	
	return partitiontype;
}

uint32_t vfat_readDirEntry(uint32_t dir_cluster, uint8_t entry_count, uint32_t *size, uint8_t *attrib, char *filename)
{
	direntry_t *dir; //Zeiger auf einen Verzeichniseintrag
	delfn_t *delfn; //Zeiger auf langen Verzeichnisseintrag
	
	uint32_t dirsector = 0;
		
	if (dir_cluster == 0)
		dirsector = RootDirectoryRegionStart;
	else
		dirsector = 0;
		
	uint8_t entry, i, j;
	signed char seqnum = -1;
	
	while (1)
	{	
		if (dir_cluster == 0)
		{//Reading from root dir
			if ((dirsector - RootDirectoryRegionStart) == RootDirectoryRegionSize)
				return 0xFFFFFFFF; //End of root dir region reached!
				
			mmc_start_read_block(dirsector++, 9);
			mmc_read_buffer(fat_buf, 512);
			mmc_stop_read_block();
		}
		else
		{
			if (vfat_readfilesector(dir_cluster, dirsector++) == 0)
				return 0xFFFFFFFF;
		}

		dir = (direntry_t *) fat_buf;
		
		for (entry=0; entry<16; entry++)
		{
			if (dir->name[0] == 0)
				return 0xFFFFFFFF;
				
			if (dir->attr == 0x0F)
			{//Entry for Long Filename
				delfn = (delfn_t *) dir;
				seqnum = (delfn->seqNumber & 0x3F) - 1;
				
				if ((entry_count == 0) && (delfn->seqNumber != 0xE5))
				{
					filename[seqnum * 13 + 0] = delfn->name1[0];
					filename[seqnum * 13 + 1] = delfn->name1[2];
					filename[seqnum * 13 + 2] = delfn->name1[4];
					filename[seqnum * 13 + 3] = delfn->name1[6];
					filename[seqnum * 13 + 4] = delfn->name1[8];
					filename[seqnum * 13 + 5] = delfn->name2[0];
					filename[seqnum * 13 + 6] = delfn->name2[2];
					filename[seqnum * 13 + 7] = delfn->name2[4];
					filename[seqnum * 13 + 8] = delfn->name2[6];
					filename[seqnum * 13 + 9] = delfn->name2[8];
					filename[seqnum * 13 + 10] = delfn->name2[10];
					filename[seqnum * 13 + 11] = delfn->name3[0];
					filename[seqnum * 13 + 12] = delfn->name3[2];
					if ((delfn->seqNumber & 0xC0))
						filename[seqnum * 13 + 13] = 0;
				}
			}
			else
			{
				if (dir->name[0] != 0xE5)
				{
					if (entry_count == 0)
					{
						*size = dir->filesize;
						*attrib = dir->attr;
					
						if (seqnum == -1)
						{//not a long filename
							for (i=0; i<8; i++)
								if (dir->name[i] == 32)
									break;
								else
									filename[i] = dir->name[i];
							
							if (dir->name[8] != 32)
							{
								filename[i] = '.';
							
								for (j=0; j<3; j++)
									if (dir->name[j + 8] == 32)
										break;
									else
										filename[i + j + 1] = dir->name[j + 8];
								
								filename[i + j + 1] = 0;
							}
							else
								filename[i] = 0;
						}
						
						if (partitiontype == PARTTYPE_FAT32)
							return dir->fstclust | ((uint32_t) dir->eaindex << 16);
						else
							return dir->fstclust;
					}
					else
						entry_count--;
				}		
				seqnum = -1;
			}	
			dir++;
		}
	}
}

static void loadfatsector(uint32_t fatpointer)
{
	if (currentfatsector != (fatpointer / 512))
	{
		mmc_start_read_block(FATRegionStart + fatpointer / 512, 9);
		mmc_read_buffer(fat_buf, 512);
		mmc_stop_read_block();
		currentfatsector = fatpointer / 512;
	}
}

uint8_t *vfat_readfilesector(uint32_t startcluster, uint32_t filesector)
{
	uint32_t clusteroffset;
	uint32_t fatpointer;
	uint32_t clusterstatus;
	uint8_t clstnum1, clstnum2;
	
	clusteroffset = filesector / SectorsPerCluster;
	filesector %= SectorsPerCluster;

		
	if (clusteroffset)
	{
		currentfatsector = 0xFFFFFFFF;
		switch (partitiontype)
		{
			case PARTTYPE_FAT12:
				while (clusteroffset)
				{
					fatpointer = startcluster * 3 / 2;
					loadfatsector(fatpointer);
					clstnum1 = fat_buf[fatpointer % 512];
					fatpointer++;
					loadfatsector(fatpointer);
					clstnum2 = fat_buf[fatpointer % 512];
					
					if ((startcluster & 0x01))
						startcluster = ((clstnum1 & 0xF0) >> 4) | (clstnum2 << 4);
					else
						startcluster = clstnum1 | ((clstnum2 & 0x0F) << 8);
					
					clusterstatus = startcluster | 0xFFFFF000;
					clusteroffset--;
				}
				break;
				
			case PARTTYPE_FAT16L:
				while (clusteroffset)
				{
					fatpointer = startcluster * 2;
					loadfatsector(fatpointer);
					startcluster = fat_buf[fatpointer % 512];
					startcluster |= fat_buf[(fatpointer + 1) % 512] << 8;
					clusterstatus = startcluster | 0xFFFF0000;
					clusteroffset--;
				}
			
			case PARTTYPE_FAT32:
				while (clusteroffset)
				{
					fatpointer = startcluster * 4;
					loadfatsector(fatpointer);
					startcluster = fat_buf[fatpointer % 512];
					startcluster |= fat_buf[(fatpointer + 1) % 512] << 8;
					startcluster |= (uint32_t) fat_buf[(fatpointer + 2) % 512] << 16;
					startcluster |= (uint32_t) fat_buf[(fatpointer + 3) % 512] << 24;
					clusteroffset--;
					clusterstatus = startcluster | 0xF0000000;
				}
		}
	}
	
	if ((startcluster == 0) | (startcluster == 1))
		return 0; //Free/reserved Sector
	
	if ((startcluster & 0xFFFFFFF0) == 0xFFFFFFF0)
		return 0; //Reserved / bad / last cluster
		
	mmc_start_read_block((startcluster - 2) * SectorsPerCluster + DataRegionStart + filesector, 9);
	mmc_read_buffer(fat_buf, 512);
	mmc_stop_read_block();
	return fat_buf;
}
